/*------------------------------------------------------------------------------
* sbsdump.c : dump sbas messages
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 22:03:15 $
* history : 2007/10/08 1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: sbsdump.c,v 1.1 2008/07/17 22:03:15 ttaka Exp $";

/* help text -----------------------------------------------------------------*/
static const char *help[]={
"",
" usage: sbsdump file",
"",
" Dump SBAS messages. Specify SBAS log file. Options are as follows.",
"",
" -h        print help",
" -b        SBAS satellite PRN number [129]",
" -s        corrected satellite PRN number [all]",
" -f        dump fast correction messages [off]",
" -i        dump ionospheric correction messages [off]",
" -l        dump long term correction messages [off]",
" -n        dump GEO navigation message [off]",
" -g        dump ionospheric grid points [off]",
" -t        dump integrity messages [off]",
};
/* print help ----------------------------------------------------------------*/
static void printhelp(void)
{
    int i;
    for (i=0;i<sizeof(help)/sizeof(*help);i++) fprintf(stderr,"%s\n",help[i]);
    exit(0);
}
/* dump sbas satellite fast corrections --------------------------------------*/
static void dumpsatf(const sbssat_t *sbssat, int sat, int type)
{
    int i,udre;
    char s[64];
    for (i=0;i<sbssat->nprn;i++) {
        if (sat!=0&&sat!=sbssat->sat[i].prn) continue;
        if (type!=0&&i/13!=type-2) continue;
        time2str(sbssat->sat[i].fcorr.t0,s,0);
        if ((udre=sbssat->sat[i].fcorr.udre-1)<0) continue;
        printf("%s : PRN%3d PRC=%7.3f IODF=%d UDRE=%2d\n",s,sbssat->sat[i].prn,
               sbssat->sat[i].fcorr.prc,sbssat->sat[i].fcorr.iodf,udre);
    }
}
/* dump sbas satellite long term corrections ---------------------------------*/
static void dumpsatl(const sbssat_t *sbssat, int sat)
{
    int i,prn;
    char s[64];
    for (i=0;i<sbssat->nprn;i++) {
        prn=sbssat->sat[i].prn;
        if (sat!=0&&sat!=prn) continue;
        time2str(sbssat->sat[i].lcorr.t0,s,0);
        printf("%s : PRN%3d IOD=%3d DP=%7.3f %7.3f %7.3f DV=%8.5f %8.5f %8.5f DC=%8.3f %8.3f\n",
               s,sbssat->sat[i].prn,sbssat->sat[i].lcorr.iode,
               sbssat->sat[i].lcorr.dpos[0],sbssat->sat[i].lcorr.dpos[1],
               sbssat->sat[i].lcorr.dpos[2],sbssat->sat[i].lcorr.dvel[0],
               sbssat->sat[i].lcorr.dvel[1],sbssat->sat[i].lcorr.dvel[2],
               sbssat->sat[i].lcorr.daf0*CLIGHT,sbssat->sat[i].lcorr.daf1*CLIGHT);
    }
}
/* dump sbas ionospheric correction ------------------------------------------*/
static void dumpionc(const sbsion_t *sbsion, int band, int block)
{
    int i,j;
    char s[64];
    for (i=0;i<15;i++) {
        if ((j=block*15+i)>=sbsion->nigp) continue;
        if (i==0) {
            time2str(sbsion->igp[j].t0,s,0);
            printf("%s : BLK=%d,%d DI=",s,band,block);
        }
        printf(" %4.2f",sbsion->igp[j].delay);
    }
    printf("\n");
}
/* dump sbas ionospheric delays ----------------------------------------------*/
static void dumpion(const sbsion_t *sbsion, int band)
{
    int i,j,give;
    for (i=0;i<=10;i++) {
        if (band!=0&&band!=i) continue;
        for (j=0;j<sbsion[i].nigp;j++) {
            give=sbsion[i].igp[j].give-1;
            printf("%2d %4d %4d %7.3f %2d\n",i,
                   sbsion[i].igp[j].lat,sbsion[i].igp[j].lon,
                   sbsion[i].igp[j].delay,give<0?15:give);
        }
    }
}
/* dump sbas igp positions ---------------------------------------------------*/
static void dumpigp(void)
{
    int i,j,k;
    for (i=0;i<=8;i++) {
        for (j=0;j<8;j++) {
            for (k=igpband1[i][j].bits;k<=igpband1[i][j].bite;k++)
                printf("%2d %4d %4d\n",i,igpband1[i][j].y[k-igpband1[i][j].bits],
                       igpband1[i][j].x);
        }
    }
    for (i=9;i<=10;i++) {
        for (j=0;j<5;j++) {
            for (k=igpband2[i-9][j].bits;k<=igpband2[i-9][j].bite;k++)
                printf("%2d %4d %4d\n",
                i,igpband2[i-9][j].x,igpband2[i-9][j].y[k-igpband2[i-9][j].bits]);
        }
    }
}
/* dump sbas navigation messages ---------------------------------------------*/
extern void dumpnav(const sbsmsg_t *msg, const sbsnav_t *sbsnav)
{
    char s1[32],s2[32];
    time2str(gpst2time(msg->week,msg->tow),s1,0);
    time2str(sbsnav->t0,s2,0);
    printf("%s : PRN%3d URA=%2d T0=%s P=%12.2f %12.2f %12.2f V=%9.5f %9.5f %9.5f A=%10.3E %10.3E %10.3E C=%10.3E %10.3E\n",
           s1,msg->prn,sbsnav->ura,s2,sbsnav->pos[0],sbsnav->pos[1],sbsnav->pos[2],
           sbsnav->vel[0],sbsnav->vel[1],sbsnav->vel[2],sbsnav->acc[0],
           sbsnav->acc[1],sbsnav->acc[2],sbsnav->af0,sbsnav->af1);
}
/* main ----------------------------------------------------------------------*/
int main(int argc, char **argv)
{
    int i,j,nmsg,dump[5]={0},ssat=0,sprn=0,type,mask,prn,band,block;
    char *file="";
    FILE *fp=stdin;
    sbsmsg_t *sbsmsg=NULL;
    sbsstat_t stat={{0}};
    
    for (i=1;i<argc;i++) {
        if      (!strcmp(argv[i],"-b")&&i+1<argc) ssat=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-s")&&i+1<argc) sprn=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-f")) dump[0]=1;
        else if (!strcmp(argv[i],"-i")) dump[3]=1;
        else if (!strcmp(argv[i],"-l")) dump[1]=1;
        else if (!strcmp(argv[i],"-n")) dump[4]=1;
        else if (!strcmp(argv[i],"-g")) dump[2]=1;
        else if (!strcmp(argv[i],"-t")) dump[5]=1;
        else if (!strncmp(argv[i],"-",1)) printhelp();
        else file=argv[i];
    }
    if (file[0]!='\0'&&!(fp=fopen(file,"r"))) {
        fprintf(stderr,"input file open error : %s\n",file);
        return -1;
    }
    nmsg=sbsreadmsg(file,ssat,&sbsmsg);
    
    for (i=0;i<nmsg;i++) {
        sbsupdatestat(sbsmsg+i,&stat);
        type=getbitu(sbsmsg[i].msg,8,6);
        switch (type) {
        case 0: case 2: case 3: case 4: case 5:
            if (!dump[0]) break;
            dumpsatf(&stat.sbssat,sprn,type);
            break;
        case 6:
            if (!dump[5]) break;
            for (j=2;j<=5;j++) {
                dumpsatf(&stat.sbssat,sprn,j);
            }
            break;
        case 9:
            if (!dump[4]) break;
            if (ssat==0||ssat==sbsmsg[i].prn)
                for (j=0;j<MAXNGEO;j++) {
                    if (stat.sbsnav[j].prn!=sbsmsg[i].prn) continue;
                    dumpnav(sbsmsg+i,stat.sbsnav+j);
                }
            break;
        case 25:
            if (!dump[1]) break;
            for (j=14;j<=120;j+=106) {
                if (getbitu(sbsmsg[i].msg,j,1)) {
                    mask=getbitu(sbsmsg[i].msg,j+1,6);
                    if (mask<=0||stat.sbssat.nprn<mask) break;
                    prn=stat.sbssat.sat[mask-1].prn;
                    if (sprn==0||sprn==prn) dumpsatl(&stat.sbssat,prn);
                }
                else {
                    mask=getbitu(sbsmsg[i].msg,j+1,6);
                    if (mask<=0||stat.sbssat.nprn<mask) break;
                    prn=stat.sbssat.sat[mask-1].prn;
                    if (sprn==0||sprn==prn) dumpsatl(&stat.sbssat,prn);
                    mask=getbitu(sbsmsg[i].msg,j+52,6);
                    if (mask<=0||stat.sbssat.nprn<mask) break;
                    prn=stat.sbssat.sat[mask-1].prn;
                    if (sprn==0||sprn==prn) dumpsatl(&stat.sbssat,prn);
                }
            }
            break;
        case 26:
            if (!dump[3]) break;
            band=getbitu(sbsmsg[i].msg,14,4);
            block=getbitu(sbsmsg[i].msg,18,4);
            dumpionc(stat.sbsion+band,band,block);
            break;
        }
    }
    if (dump[2]) dumpigp();
    fclose(fp);
    return 0;
}
