/*------------------------------------------------------------------------------
* rtklib.h : RTKLIB constants, type definitions and function prototypes
*
*          Copyright (C) 2006-2009 by T.TAKASU, All rights reserved.
*
* options : -DENAGLO   enable GLONASS
*           -DENAGAL   enable Galileo
*           -DENAQZS   enable QZSS
*           -DNFREQ=n  set number of carrier frequencies
*           -DMAXOBS=n set max number of obs data in an epoch
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2007/01/13 1.0  RTKLIB ver.1.0.0
*           2007/03/20 1.1  RTKLIB ver.1.1.0
*           2008/07/15 1.2  RTKLIB ver.2.1.0
*           2008/10/19 1.3  RTKLIB ver.2.1.1
*           2009/01/31 1.4  RTKLIB ver.2.2.0
*           2009/04/30 1.5  RTKLIB ver.2.2.1
*-----------------------------------------------------------------------------*/
#ifndef RTKLIB_H
#define RTKLIB_H
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#ifdef WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif
#ifdef __cplusplus
extern "C" {
#endif

/* constants -----------------------------------------------------------------*/

#define VER_RTKLIB  "2.2.1"             /* library version */

#define PI          3.1415926535897932  /* pi */
#define D2R         (PI/180.0)          /* deg to rad */
#define R2D         (180.0/PI)          /* rad to deg */
#define CLIGHT      299792458.0         /* speed of light (m/s) */
#define SC2RAD      3.1415926535898     /* semi-circle to radian (IS-GPS) */
#define MU          3.986005E14         /* earth gravitational constant (IS-GPS) */
#define OMGE        7.2921151467E-5     /* earth angular velocity (IS-GPS) (rad/s) */
#define RE          6378137.0           /* earth semimajor axis (WGS84) (m) */
#define FE          (1.0/298.257223563) /* earth flattening (WGS84) */
#define HION        350000.0            /* ionosphere height (m) */

#define FREQ1       1.57542E9           /* L1/E1  frequency (Hz) */
#define FREQ2       1.22760E9           /* L2     frequency (Hz) */
#define FREQ5       1.17645E9           /* L5/E5a frequency (Hz) */
#define FREQE5b     1.20714E9           /* E5b    frequency (Hz) */
#define FREQE6      1.27875E9           /* E6/LEX frequency (Hz) */

#ifndef NFREQ
#define NFREQ       2                   /* number of carrier frequencies */
#endif
#define MAXPRNGPS   32                  /* max satellite PRN number of GPS */
#ifdef ENAGLO
#define MAXPRNGLO   24                  /* max satellite slot number of GLONASS */
#else
#define MAXPRNGLO   0
#endif
#ifdef ENAGAL
#define MAXPRNGAL   27                  /* max satellite PRN number of Galileo */
#else
#define MAXPRNGAL   0
#endif
#ifdef ENAQZS
#define MAXPRNQZS   3                   /* max satellite sat number of QZSS */
#else
#define MAXPRNQZS   0
#endif

#define MINPRNSBS   120                 /* min satellite PRN number of SBAS */
#define MAXPRNSBS   138                 /* max satellite PRN number of SBAS */
#define MAXSAT (MAXPRNGPS+MAXPRNGLO+MAXPRNGAL+MAXPRNQZS+MAXPRNSBS-MINPRNSBS+1)
                                        /* max satellite number (1 to MAXSAT) */
#ifndef MAXOBS
#define MAXOBS      64                  /* max number of obs in an epoch */
#endif
#define MAXRCV      64                  /* max receiver numver (1 to MAXRCV) */
#define MAXOBSTYPE  32                  /* max number of obs type in rinex */
#define DTTOL       0.02                /* tolerance of time difference (s) */
#define MAXDTOE     7200.0              /* max time difference to ephem toe (s) for gps */
#define MAXDTOE_GLO 900.0               /* max time difference to glonas toe (s) */
#define MAXDTOE_S   86400.0             /* max time difference to ephem toe (s) for other */
#define MAXGDOP     300.0               /* max GDOP */

#define MAXEXFILE   100                 /* max number of expanded files */
#define MAXSBSAGEF  30.0                /* max age of SBAS fast correction (s) */
#define MAXSBSAGEL  1800.0              /* max age of SBAS long term corr (s) */
#define MAXSBSAGEN  360.0               /* max age of SBAS navigation message (s) */
#define MAXSBSURA   8                   /* max URA of SBAS satellite */
#define MAXBAND     10                  /* max SBAS band of igp */
#define MAXNIGP     201                 /* max number of IGP in SBAS band */
#define MAXNGEO     4                   /* max number of GEO satellites */
#define MAXCOMMENT  10                  /* max number of RINEX comments */
#define MAXSTRPATH  1024                /* max length of stream path */
#define MAXSTRMSG   1024                /* max length of stream message */
#define MAXSTRRTK   6                   /* max number of stream in rtk server */
#define MAXSBSMSG   32                  /* max number of sbas msg in rtk server */
#define MAXSOLMSG   4096                /* max length of solution message */
#define MAXRAWLEN   4096                /* max length of receiver raw message */
#define MAXERRMSG   4096                /* max length of error/warning message */
#define MAXSOLBUF   256                 /* max nuber of rtk solution buffer */

#define SYS_NONE    0x00                /* navigation system: none */
#define SYS_GPS     0x01                /* navigation system: GPS */
#define SYS_SBS     0x02                /* navigation system: SBAS */
#define SYS_GLO     0x04                /* navigation system: GLONASS */
#define SYS_GAL     0x08                /* navigation system: Galileo */
#define SYS_QZS     0x10                /* navigation system: QZSS */
#define SYS_ALL     0xFF                /* navigation system: all */

#define OBSTYPE_PR  0x01                /* observation type: pseudorange */
#define OBSTYPE_CP  0x02                /* observation type: carrier-phase */
#define OBSTYPE_DOP 0x04                /* observation type: doppler-freq. */
#define OBSTYPE_SNR 0x08                /* observation type: SNR */
#define OBSTYPE_ALL 0xFF                /* observation type: all */

#define FREQTYPE_L1 0x01                /* frequency type: L1/E1 */
#define FREQTYPE_L2 0x02                /* frequency type: L2 */
#define FREQTYPE_L5 0x04                /* frequency type: L5/E5a */
#define FREQTYPE_L7 0x08                /* frequency type: E5b */
#define FREQTYPE_L6 0x10                /* frequency type: E6/LEX */
#define FREQTYPE_ALL 0xFF               /* frequency type: all*/

#define PMODE_SINGLE 0                  /* positioning mode: single */
#define PMODE_DGPS   1                  /* positioning mode: dgps */
#define PMODE_KINEMA 2                  /* positioning mode: kinematic */
#define PMODE_STATIC 3                  /* positioning mode: static */
#define PMODE_MOVEB  4                  /* positioning mode: moving-base */
#define PMODE_FIXED  5                  /* positioning mode: fixed */

#define SOLF_LLH    0                   /* solution format: lat/lon/height */
#define SOLF_XYZ    1                   /* solution format: x/y/z-ecef */
#define SOLF_ENU    2                   /* solution format: e/n/u-baseline */
#define SOLF_NMEA   3                   /* solution format: nmea-183 */

#define SOLQ_NONE   0                   /* solution status: no solution */
#define SOLQ_FIX    1                   /* solution status: fix */
#define SOLQ_FLOAT  2                   /* solution status: float */
#define SOLQ_SBAS   3                   /* solution status: SBAS */
#define SOLQ_DGPS   4                   /* solution status: DGPS */
#define SOLQ_SINGLE 5                   /* solution status: single */

#define TIMES_GPST  0                   /* time system: gps time */
#define TIMES_UTC   1                   /* time system: UTC */
#define TIMES_JST   2                   /* time system: JST */

#define SBSOPT_LCORR 1                  /* SBAS option: long term correction */
#define SBSOPT_FCORR 2                  /* SBAS option: fast correction */
#define SBSOPT_ICORR 4                  /* SBAS option: ionosphere correction */
#define SBSOPT_RANGE 8                  /* SBAS option: ranging */

#define STR_NONE     0                  /* stream type: none */
#define STR_SERIAL   1                  /* stream type: serial */
#define STR_FILE     2                  /* stream type: file */
#define STR_TCPSVR   3                  /* stream type: tcp server */
#define STR_TCPCLI   4                  /* stream type: tcp client */
#define STR_UDP      5                  /* stream type: udp stream */
#define STR_NTRIPSVR 6                  /* stream type: ntrip server */
#define STR_NTRIPCLI 7                  /* stream type: ntrip client */

#define STRFMT_RTCM2 0                  /* stream format: RTCM 2 */
#define STRFMT_RTCM3 1                  /* stream format: RTCM 3 */
#define STRFMT_OEM4  2                  /* stream format: NovAtel OEMV/4 raw */
#define STRFMT_OEM3  3                  /* stream format: NovAtel OEM3 raw */
#define STRFMT_UBX   4                  /* stream format: u-blox raw */
#define STRFMT_SS2   5                  /* stream format: Superstar II raw */
#define STRFMT_CRES  6                  /* stream format: Crescent raw */
#define STRFMT_GARM  7                  /* stream format: Garmin raw */
#define STRFMT_SIRF  8                  /* stream format: SiRF II raw */
#define STRFMT_SBAS  10                 /* stream format: SBAS messages */
#define STRFMT_NMEA  11                 /* stream format: NMEA */
#define STRFMT_RINEX 12                 /* stream format: RINEX */

#define STR_MODE_R  0x1                 /* stream mode: read */
#define STR_MODE_W  0x2                 /* stream mode: write */

#define P2_5        0.03125             /* 2^-5 */
#define P2_11       4.882812500000000E-04 /* 2^-11 */
#define P2_19       1.907348632812500E-06 /* 2^-19 */
#define P2_24       5.960464477539063E-08 /* 2^-24 */
#define P2_27       7.450580596923828E-09 /* 2^-27 */
#define P2_29       1.862645149230957E-09 /* 2^-29 */
#define P2_30       9.313225746154785E-10 /* 2^-30 */
#define P2_31       4.656612873077393E-10 /* 2^-31 */
#define P2_33       1.164153218269348E-10 /* 2^-33 */
#define P2_39       1.818989403545856E-12 /* 2^-39 */
#define P2_43       1.136868377216160E-13 /* 2^-43 */
#define P2_50       8.881784197001252E-16 /* 2^-50 */
#define P2_55       2.775557561562891E-17 /* 2^-55 */

#ifdef WIN32
#define thread_t    HANDLE
#define lock_t      CRITICAL_SECTION
#define initlock(f) InitializeCriticalSection(f)
#define lock(f)     EnterCriticalSection(f)
#define unlock(f)   LeaveCriticalSection(f)
#else
#define thread_t    pthread_t
#define lock_t      pthread_mutex_t
#define initlock(f) pthread_mutex_init(f,NULL)
#define lock(f)     pthread_mutex_lock(f)
#define unlock(f)   pthread_mutex_unlock(f)
#endif

/* type definitions ----------------------------------------------------------*/

typedef struct {        /* time struct */
    time_t time;        /* time (s) expressed by standard time_t */
    double sec;         /* fraction of sec under 1 s */
} gtime_t;

typedef struct {        /* observation data record */
    gtime_t time;       /* receiver sampling time */
    unsigned char sat,rcv;    /* satellite/receiver number */
    unsigned char SNR[NFREQ]; /* signal strength (dbHz) */
    unsigned char LLI[NFREQ]; /* loss of lock indicator */
    double L[NFREQ];    /* observation data carrier-phase (cycle) */
    double P[NFREQ];    /* observation data pseudorange (m) */
    float  D[NFREQ];    /* observation data doppler frequency (Hz) */
} obsd_t;

typedef struct {        /* observation data */
    int n;              /* number of obervation data */
    int nmax;           /* number of allocated obervation data */
    obsd_t *data;       /* observation data records */
    double pos [MAXRCV][3]; /* approx position x/y/z (m) */
    double adel[MAXRCV][3]; /* antenna delta e/n/u (m) */
} obs_t;

typedef struct {        /* satellite broadcast ephemeris */
    int sat;            /* satellite number */
    int iode,iodc;      /* IODE,IODC */
    int sva,svh;        /* sv accuracy (ura index), sv health */
    int week;           /* gps week number */
    int l2code,l2pflg;  /* code on L2, L2 P data flag */
    gtime_t toe,toc,ttr; /* Toe,Toc,T_trans */
    double A,e,i0,OMG0,omg,M0,deln,OMGd,idot;
    double crc,crs,cuc,cus,cic,cis,toes,fit;
                        /* sv ephemeris parameters */
    double f0,f1,f2,tgd; /* sv clock parameters */
} eph_t;

typedef struct {        /* glonass broadcast ephemeris */
    int sat;            /* satellite number */
    int frq;            /* satellite frequency number */
    int svh,age;        /* satellite health, age of operation */
    gtime_t toe;        /* epoch of epherides (gpst) */
    gtime_t tof;        /* message frame time (gpst) */
    double pos[3];      /* satellite position (ecef) (m) */
    double vel[3];      /* satellite velocity (ecef) (m/s) */
    double acc[3];      /* satellite acceleration (ecef) (m/s^2) */
    double taun,gamn;   /* sv clock bias (s)/relative freq bias */
} geph_t;

typedef struct {        /* satellite precise ephemeris */
    gtime_t time;       /* time */
    double pos[MAXSAT][4]; /* satellite position/clock (ecef) (m,s) */
    double std[MAXSAT][4]; /* satellite position/clock std (m,s) */
} peph_t;

typedef struct {        /* navigation messages */
    int n;              /* number of broadcast ephemeris */
    int nmax;           /* number of allocated broadcast ephemeris */
    int ng;             /* number of glonass ephemeris */
    int ngmax;          /* number of allocated glonass ephemeris */
    int ne;             /* number of precise ephemeris */
    int nemax;          /* number of allocated precise ephemeris */
    eph_t  *eph;        /* broadcast ephemeris */
    geph_t *geph;       /* glonass broadcast ephemeris */
    peph_t *peph;       /* precise ephemeris */
    double sap[MAXSAT][3]; /* sat antenna phase center position (m) */
    double ion[8];      /* iono model parameters {a0,a1,a2,a3,b0,b1,b2,b3} */
    double utc[4];      /* delta-utc parameters */
    int leaps;          /* leap seconds (s) */
} nav_t;

typedef struct {        /* sbas message */
    int week,tow;       /* receiption time */
    int prn;            /* sbas satellite prn number */
    unsigned char msg[29]; /* sbas message (226bit) padded by 0 */
} sbsmsg_t;

typedef struct {        /* sbas messages */
    int n;              /* number of sbas messages */
    sbsmsg_t *msgs;     /* sbas messages */
} sbs_t;

typedef struct {        /* sbas fast correction type */
    gtime_t t0;         /* time of applicability (tof) */
    double prc;         /* pseudorange correction (prc) (m) */
    double rrc;         /* range-rate correction (rrc) (m/s) */
    double dt;          /* range-rate correction delta-time (s) */
    int iodf;           /* iodf (issue of date fast corr) */
    short udre;         /* udre+1 */
    short ai;           /* degradation factor indicator */
} sbsfcorr_t;

typedef struct {        /* sbas long term satellite error correction type */
    gtime_t t0;         /* correction time */
    int iode;           /* iode (issue of date ephemeris) */
    double dpos[3];     /* delta position (m) (ecef) */
    double dvel[3];     /* delta velocity (m/s) (ecef) */
    double daf0,daf1;   /* delta clock-offset/drift (s,s/s) */
} sbslcorr_t;
 
typedef struct {        /* sbas satellite correction type */
    int prn;            /* satellite prn number */
    sbsfcorr_t fcorr;   /* fast correction */
    sbslcorr_t lcorr;   /* long term correction */
} sbssatp_t;

typedef struct {        /* sbas satellite corrections type */
    int iodp;           /* iodp (issue of date mask) */
    int nprn;           /* number of satellites */
    int tlat;           /* system latency (s) */
    sbssatp_t sat[MAXSAT]; /* satellite correction */
} sbssat_t;

typedef struct {        /* sbas ionospheric correction type */
    gtime_t t0;         /* correction time */
    short lat,lon;      /* latitude/longitude (deg) */
    short give;         /* give+1 */
    float delay;        /* vertical delay estimate (m) */
} sbsigp_t;

typedef struct {        /* igp band type */
    short x;            /* longitude/latitude (deg) */
    const short *y;     /* latitudes/longitudes (deg) */
    unsigned char bits; /* igp mask start bit */
    unsigned char bite; /* igp mask end bit */
} sbsigpband_t;

typedef struct {        /* sbas ionospheric corrections type */
    int iodi;           /* iodi (issue of date ionos corr) */
    int nigp;           /* number of igps */
    sbsigp_t igp[MAXNIGP]; /* ionospheric correction */
} sbsion_t;

typedef struct {        /* sbas navigation message type */
    int prn;            /* satellite prn number */
    gtime_t t0;         /* reference epoch time */
    int ura;            /* ura+1 (user range accuracy) */
    double pos[3];      /* satellite position (m) (ecef) */
    double vel[3];      /* satellite velocity (m/s) (ecef) */
    double acc[3];      /* satellite acceleration (m/s^2) (ecef) */
    double af0,af1;     /* satellite clock-offset/drift (s,s/s) */
} sbsnav_t;

typedef struct {        /* sbas status type */
    sbssat_t sbssat;    /* satellite corrections */
    sbsion_t sbsion[MAXBAND+1]; /* ionos corrections */
    sbsnav_t sbsnav[MAXNGEO];   /* sbas navigation messages */
} sbsstat_t;

typedef struct {        /* antenna phase center parameters */
    double ecc[3];      /* antenna delta-position e/n/u (m) */
    double off[NFREQ][ 3]; /* phase center offset e/n/u (m) */
    double var[NFREQ][19]; /* phase center variation (m) (el=0,5,...,90deg) */
    char ant[128];      /* antenna name */
} pcv_t;

typedef struct {        /* rinex options */
    gtime_t ts,te;      /* time start/end */
    double tint;        /* time interval (s) */
    int navsys;         /* navigation system */
    int obstype;        /* observation type */
    int freqtype;       /* frequency type */
    int opt;            /* receiver dependent options */
    char prog  [32];    /* program */
    char runby [32];    /* run-by */
    char marker[64];    /* marker name */
    char name[2][32];   /* observer/agency */
    char rec [3][32];   /* receiver #/type/vers */
    char ant [3][32];   /* antenna #/type */
    double apppos[3];   /* approx position x/y/z */
    double antdel[3];   /* antenna delta h/e/n */
    char comment[MAXCOMMENT][64]; /* comments */
    gtime_t tstart;     /* first obs time */
    gtime_t tend;       /* last obs time */
    gtime_t trtcm;      /* approx log start time for rtcm */
} rnxopt_t;

typedef struct {        /* position solution data type */
    gtime_t time;       /* solution time */
    double pos[3];      /* receiver position lat/lon/height (deg/m) */
    float poss[6];      /* receiver position covariance {ee,nn,uu,en,nu,ue} */
    short posq,ns;      /* solution status(SOLQ_???)/number of satellites */
} solpd_t;

typedef struct {        /* position solution type */
    int n;              /* number of solution data */
    solpd_t *data;      /* position solution data */
} solp_t;

typedef struct {        /* rtk solution type */
    int stat;           /* status (0:inactive,1:fix,2:float,4:dgps,5:single) */
    gtime_t time;       /* time of rover */
    double rr[6];       /* rover position/velocity (ecef) (m,m/s) */
    double Qr[9];       /* rover position covariance (3x3) */
    double rb[3];       /* base-station position (ecef) (m) */
    int ns;             /* number of valid satellite */
    double age;         /* age of differential */
    double ratio;       /* ar valiation ratio */
} rtksol_t;

typedef struct {        /* reference station parameter type */
    double rr[3];       /* station position (ecef) (m) */
    double del[NFREQ][3]; /* antenna position delta (ecef) (m) */
    double anth;        /* antenna height (m) */
    int itrf;           /* itrf realization year */
} rtcmsta_t;

typedef struct {        /* DGPS correction type */
    gtime_t t0;         /* correction time */
    double prc;         /* pseudorange correction (PRC) (m) */
    double rrc;         /* range rate correction (RRC) (m/s) */
    int iod;            /* issue of data */
    double udre;        /* UDRE */
} rtcmdgps_t;

typedef struct {        /* RTCM control struct type */
    int staid;          /* station id */
    int stah;           /* station health */
    int seqno;          /* sequence number for rtcm 2 */
    gtime_t time;       /* message time */
    obs_t obs;          /* observation data (uncorrected) */
    nav_t nav;          /* satellite ephemerides */
    rtcmsta_t sta;      /* reference station parameter */
    rtcmdgps_t dgps[MAXSAT]; /* DGPS correction */
    char msg[128];      /* special message */
    int obsflag;        /* obs data complete flag (1:ok,0:not complete) */
    int ephsat;         /* sat number of update ephemeris (0:no satellite) */
    double cp[MAXSAT][NFREQ]; /* carrier-phase measurement */
    unsigned char lock[MAXSAT][NFREQ]; /* lock time */
    unsigned char loss[MAXSAT][NFREQ]; /* loss of lock count */
    int nbyte;          /* number of bytes in message buffer */ 
    int nbit;           /* number of bits in word buffer */ 
    int len;            /* message length (bytes) */
    unsigned char buff[1200]; /* message buffer */
    unsigned int word;  /* word buffer for rtcm 2 */
    int nmsg2[64];      /* message count of rtcm 2 (1-63:1-63,0:other) */
    int nmsg3[64];      /* message count of rtcm 3 (1-63:1001-1063,0:ohter) */
} rtcm_t;

typedef struct {        /* processing options type */
    int mode;           /* positioning mode (PMODE_???) */
    int nf;             /* number of frequencies (1:L1,2:L1+L2,3:L1+L2+L5) */
    double elmin;       /* elevation mask angle (rad) */
    double snrmin;      /* snr mask (dBHz) */
    int sateph;         /* satellite ephemeris (0:broadcast,1:precise) */
    int modear;         /* AR mode (0:off,1:continuous,2:instantaneous) */
    int maxout;         /* obs outage count to reset bias */
    int minlock;        /* min lock count to fix ambiguity */
    int estiono;        /* ionosphere estimation */
    int esttrop;        /* troposphere estimation */
    int niter;          /* number of filter iteration */
    int codesmooth;     /* code smoothing window size (0:none) */
    int intpref;        /* interpolate reference obs (for post mission) */
    int sbascorr;       /* SBAS correction options */
    int refpos;         /* base station position */
                        /* (0:pos in proc option,1: average of single pos, */
                        /*  2:rinex header, 3:rtcm pos) */
    double err[4];      /* measurement error factor */
                        /* [0]:code/phase, [1-3]:error factor a/b/c of phase */
    double std[3];      /* initial-state std [0]bias,[1]ionos [2]tropos */
    double prn[3];      /* process-noise std [0]bias,[1]ionos [2]tropos */
    double sclkstab;    /* satellite clock stability (sec/sec) */
    double thresar;     /* AR validation threshold */
    double elmaskar;    /* elevation mask of AR for rising satellite (deg) */
    double thresslip;   /* slip threshold of geometry-free phase (m) */
    double maxtdiff;    /* max difference of time (sec) */
    double maxinno;     /* reject threshold of innovation (m) */
    double ru[3];       /* rover position for fixed mode(5) (ecef) (m) */
    double rb[3];       /* base station position for rel mode(1-3) (ecef) (m) */
    pcv_t *pcvu;        /* antenna phase center parameters of rover */
    pcv_t *pcvr;        /* antenna phase center parameters of reference */
    unsigned char exsats[MAXSAT]; /* excluded satellites (1:excluded) */
} prcopt_t;

typedef struct {        /* solution options type */
    int posf;           /* solution format (SOLF_???) */
    int times;          /* time system (TIMES_???) */
    int timef;          /* time format (0:sssss.s,1:yyyy/mm/dd hh:mm:ss.s) */
    int timeu;          /* time digits under decimal point */
    int degf;           /* latitude/longitude format (0:ddd.ddd,1:ddd mm ss) */
    int outhead;        /* output header (0:no,1:yes) */
    int outopt;         /* output processing options (0:no,1:yes) */
    int datum;          /* datum (0:WGS84,1:Tokyo) */
    int height;         /* height (0:ellipsoidal,1:geodetic) */
    int geoid;          /* geoid model (0:EGM96,1:JGD2000) */
    char sep[64];       /* field separator */
} solopt_t;

typedef struct {        /* rtk control/result type */
    gtime_t time[2];    /* receiver time */
    int nx,na;          /* number of float states/fixed states */
    int ns;             /* number of valid satellites */
    double tt;          /* time difference between current and previous (s) */
    double age;         /* age of differential (s) */
    double ambratio;    /* raitio for ambiguity validation */
    double *x,*P;       /* float states and their covariance */
    double *xa,*Pa;     /* fixed states and their covariance */
    double rr[2][6];      /* position/velocity single/fixed (ecef) (m,m/s) */
    double az[2][MAXSAT]; /* azimuth angle (rad) */
    double el[2][MAXSAT]; /* elevation angle (rad) */
    unsigned char vs[2][MAXSAT]; /* valid satellite flag single */
    double        resp [MAXSAT][NFREQ]; /* residuals of pseudorange (m) */
    double        resc [MAXSAT][NFREQ]; /* residuals of carrier-phase (m) */
    unsigned char vsat [MAXSAT][NFREQ]; /* valid satellite flag */
    unsigned char slip [MAXSAT][NFREQ]; /* cycle-slip flag */
    unsigned int  lock [MAXSAT][NFREQ]; /* lock counter of phase */
    unsigned int  outc [MAXSAT][NFREQ]; /* obs outage counter of phase */
    unsigned int  slipc[MAXSAT][NFREQ]; /* cycle-slip counter */
    unsigned int  rejc [MAXSAT][NFREQ]; /* reject counter */
    double gf[MAXSAT];  /* geometry-free phase (m) */
    int neb;            /* bytes in error message buffer */
    char errbuf[MAXERRMSG]; /* error message buffer */
    prcopt_t opt;       /* processing options */
} rtk_t;

typedef struct {        /* receiver raw data control type */
    gtime_t time;       /* message time */
    gtime_t tobs;       /* observation data time */
    obs_t obs;          /* observation data */
    nav_t nav;          /* satellite ephemerides */
    int ephsat;         /* sat number of update ephemeris (0:no satellite) */
    sbsmsg_t sbsmsg;    /* sbas message */
    unsigned char subfrm[MAXSAT][120];  /* subframe buffer (1-4) */
    double lockt[MAXSAT][NFREQ]; /* lock time (s) */
    double icpp[MAXSAT],off[MAXSAT],icpc; /* carrier params for ss2 */
    unsigned char halfc[MAXSAT][NFREQ]; /* half-cycle add flag */
    int nbyte;          /* number of bytes in message buffer */ 
    int len;            /* message length (bytes) */
    unsigned char buff[MAXRAWLEN]; /* message buffer */
} raw_t;

typedef struct {        /* stream type */
    int type;           /* type (STR_???) */
    int mode;           /* mode (STR_MODE_?) */
    int state;          /* state (-1:error,0:close,1:open) */
    unsigned int inb,inr;   /* input bytes/rate */
    unsigned int outb,outr; /* output bytes/rate */
    unsigned int tick,tact; /* tick/active tick */
    unsigned int inbt,outbt; /* input/output bytes at tick */
    lock_t lock;        /* lock flag */
    void *port;         /* type dependent port control struct */
    char path[MAXSTRPATH]; /* stream path */
    char msg [MAXSTRMSG];  /* stream message */
} stream_t;

typedef struct {        /* stream server type */
    int state;          /* server state (0:stop,1:running) */
    int cycle;          /* server cycle (ms) */
    int buffsize;       /* input/monitor buffer size (bytes) */
    int nmeacycle;      /* nmea request cycle (ms) (0:no) */
    int nstr;           /* number of streams (1 input + (nstr-1) outputs */
    int npb;            /* data length in peek buffer (bytes) */
    double nmeapos[3];  /* nmea request position (ecef) (m) */
    unsigned char *buff; /* input buffers */
    unsigned char *pbuf; /* peek buffer */
    unsigned int tick;  /* start tick */
    stream_t stream[16]; /* input/output streams */
    thread_t thread;    /* server thread */
    lock_t lock;        /* lock flag */
} strsvr_t;

typedef struct {        /* rtk server type */
    int state;          /* server state (0:stop,1:running) */
    int cycle;          /* processing cycle (ms) */
    int nmeacycle;      /* nmea request cycle (ms) (0:no req) */
    int nmeareq;        /* nmea request (0:no,1:nmeapos,2:single sol) */
    double nmeapos[3];  /* nmea request position (ecef) (m) */
    int buffsize;       /* input buffer size (bytes) */
    int format[4];      /* input (rov/base)/output(2) format */
    int navsel;         /* navigation data select (0:rover,1:base,2:both) */
    int sbssat;         /* valid sbas prn number (0:all) */
    int nsbs;           /* number of sbas message */
    int nsol;           /* number of solution buffer */
    int rtkstat;        /* rtk status (SOLQ_???) */
    rtk_t rtk;          /* rtk control/result struct */
    int nb [2];         /* bytes in input buffers (rover/base) */
    int nsb[2];         /* bytes in soulution buffers */
    int npb[2];         /* bytes in input peek buffers */
    unsigned char *buff[2]; /* input buffers (rover/base) */
    unsigned char *sbuf[2]; /* output buffer (solution1/2) */
    unsigned char *pbuf[2]; /* peek buffers */
    rtksol_t solbuf[MAXSOLBUF]; /* rtk solution buffer */
    unsigned int nmsg[2][10]; /* input message counts */
    raw_t  raw [2];     /* receiver raw control (rover/base) */
    rtcm_t rtcm[2];     /* rtcm control (rover/base) */
    obs_t obs[2];       /* observation data (rover/base) */
    nav_t nav;          /* navigation messages */
    sbsstat_t sbsstat;  /* sbas status */
    sbsmsg_t sbsmsg[MAXSBSMSG]; /* sbas message buffer */
    stream_t stream[MAXSTRRTK]; /* input/output/log streams */
    unsigned int tick;  /* start tick */
    thread_t thread;    /* server thread */
    int cputime;        /* cpu time (ms) for a processing cycle */
    int prcout;         /* missing observation data count */
    lock_t lock;        /* lock flag */
} rtksvr_t;

/* global variables ----------------------------------------------------------*/
extern const double chisqr[];           /* chi-sqr(n) table (alpha=0.001) */
extern const double lam[];              /* carrier wave length (m) {L1,L2,...} */
extern const prcopt_t prcopt_default;   /* default positioning options */
extern const solopt_t solopt_default;   /* default solution output options */
extern const sbsigpband_t igpband1[][8]; /* sbas igp band 0-8 */
extern const sbsigpband_t igpband2[][5]; /* sbas igp band 9-10 */
extern const char *rcvnamelist[];       /* supported receiver name list */

/* satellite number/system functions -----------------------------------------*/
extern int  satno   (int sys, int prn);
extern int  satsys  (int sat, int *prn);
extern int  satid2no(const char *id);
extern void satno2id(int sat, char *id);

/* matrix and vector functions -----------------------------------------------*/
extern double *mat  (int n, int m);
extern int    *imat (int n, int m);
extern double *zeros(int n, int m);
extern double *eye  (int n);
extern double dot (const double *a, const double *b, int n);
extern double norm(const double *a, int n);
extern void matcpy(double *A, const double *B, int n, int m);
extern void matmul(const char *tr, int n, int k, int m, double alpha,
                   const double *A, const double *B, double beta, double *C);
extern int  matinv(double *A, int n);
extern int  solve (const char *tr, const double *A, const double *Y, int n,
                   int m, double *X);
extern int  lsq   (const double *A, const double *y, int n, int m, double *x,
                   double *Q);
extern int  filter(double *x, double *P, const double *H, const double *v,
                   const double *R, int n, int m);
extern int  smoother(const double *xf, const double *Qf, const double *xb,
                    const double *Qb, int n, double *xs, double *Qs);
extern void matprint (const double *A, int n, int m, int p, int q);
extern void matfprint(const double *A, int n, int m, int p, int q, FILE *fp);

/* time and string functions -------------------------------------------------*/
extern double str2num(const char *s, int i, int n);
extern int     str2time(const char *s, int i, int n, gtime_t *t);
extern void    time2str(gtime_t t, char *str, int n);
extern gtime_t epoch2time(const double *ep);
extern void    time2epoch(gtime_t t, double *ep);
extern gtime_t gpst2time(int week, double sec);
extern double  time2gpst(gtime_t t, int *week);

extern gtime_t timeadd  (gtime_t t, double sec);
extern double  timediff (gtime_t t1, gtime_t t2);
extern gtime_t gpst2utc (gtime_t t);
extern gtime_t utc2gpst (gtime_t t);
extern gtime_t timeget  (void);
extern double  time2doy (gtime_t t);

extern int adjgpsweek(int week);
extern unsigned int tickget(void);
extern void sleepms(int ms);

/* coordinates functions -----------------------------------------------------*/
extern void ecef2pos(const double *r, double *pos);
extern void pos2ecef(const double *pos, double *r);
extern void ecef2enu(const double *pos, const double *r, double *e);
extern void enu2ecef(const double *pos, const double *e, double *r);
extern void covenu  (const double *pos, const double *P, double *Q);
extern void covecef (const double *pos, const double *Q, double *P);
extern void xyz2enu (const double *pos, double *E);

/* input/output functions ----------------------------------------------------*/
extern int  readpcv(const char *file, const char *ant, pcv_t *pcv);
extern void readpos(const char *file, const char *rcv, double *pos);
extern int  sortobs(obs_t *obs);
extern void uniqeph(nav_t *nav);
extern int  screent(gtime_t time, gtime_t ts, gtime_t te, double tint);

/* debug trace functions -----------------------------------------------------*/
extern void traceopen(const char *file);
extern void traceclose(void);
extern void tracelevel(int level);
extern void trace   (int level, const char *format, ...);
extern void tracet  (int level, const char *format, ...);
extern void tracemat(int level, const double *A, int n, int m, int p, int q);
extern void traceobs(int level, const obsd_t *obs, int n);
extern void tracenav(int level, const nav_t *nav);
extern void traceb  (int level, const unsigned char *p, int n);

/* platform dependent functions ----------------------------------------------*/
extern int execcmd(const char *cmd);
extern int expath (const char *path, char *paths[], int nmax);

/* positioning models --------------------------------------------------------*/
extern void eph2pos (gtime_t time, const eph_t  *eph,  double *rs, double *dts);
extern void geph2pos(gtime_t time, const geph_t *geph, double *rs, double *dts);
extern void satpos (gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                    double *rs, double *dts);
extern void satposv(gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                    double *rs, double *dts);
extern void satposiode(gtime_t time, const obsd_t *obs, const int *iode, int n,
                       const nav_t *nav, double *rs, double *dts);
extern double satazel(const double *pos, const double *e, double *azel);
extern double geodist(const double *rs, const double *rr, double *e);
extern void   dops(int ns, const double *azel, double elmin, double *dop);
extern double ionmodel(gtime_t t, const double *ion, const double *pos,
                       const double *azel);
extern double ionmapf(const double *pos, const double *azel);
extern double tropmodel(const double *pos, const double *azel, double humi);
extern double tropmapf(gtime_t time, const double *pos, const double *azel,
                       double *mapfw);
extern void antmodel(const pcv_t *pcv, const double *azel, double *dr);
extern void csmooth(obs_t *obs, int ns);

/* single-point positioning --------------------------------------------------*/
extern int pntpos(const obsd_t *obs, int n, const nav_t *nav, double elmin,
                  double snrmin, double *rr, double *Qr, double *dtr,
                  double *azel, int *vsat);
extern int pntvel(const obsd_t *obs, int n, const nav_t *nav, const double *rr,
                  const double *azel, const int *vsat, double *vr, double *Qv,
                  double *ddtr);

/* geiod model ---------------------------------------------------------------*/
extern double geoidh(const double *pos);

/* datum transformation ------------------------------------------------------*/
extern int loaddatump(const char *file);
extern int tokyo2jgd(double *pos);
extern int jgd2tokyo(double *pos);

/* rinex functions -----------------------------------------------------------*/
extern int readrnx (char *files[], int n, obs_t *obs, nav_t *nav);
extern int readrnxt(char *files[], int n, gtime_t ts, gtime_t te, double tint,
                    obs_t *obs, nav_t *nav);
extern int outrnxobsh(FILE *fp, const rnxopt_t *opt);
extern int outrnxobsb(FILE *fp, const rnxopt_t *opt, const obsd_t *obs, int n,
                      int epflag);
extern int outrnxnavh(FILE *fp, const rnxopt_t *opt, const double *ion,
                      const double *dutc, int leaps);
extern int outrnxnavb(FILE *fp, const rnxopt_t *opt, const eph_t *eph);
extern int uncompress(const char *file, char *uncfile);
extern int convrnx(int format, const rnxopt_t *opt, const char *file, 
                   const char *ofile, const char *nfile, const char *sfile);

/* precise ephemeris functions -----------------------------------------------*/
extern int readsp3(const char *file, peph_t **peph);
extern int readsap(const char *file, gtime_t time, double sap[][3]);
extern int eph2posp(gtime_t time, int sat, peph_t *eph, int neph, double *rs,
                    double *dts);
extern void satposp(gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                    double *rs, double *dts);

/* receiver raw data functions -----------------------------------------------*/
extern unsigned int getbitu(const unsigned char *buff, int pos, int len);
extern int          getbits(const unsigned char *buff, int pos, int len);
extern unsigned int crc32  (const unsigned char *buff, int len);
extern unsigned int crc24q (const unsigned char *buff, int len);
extern int decode_word (unsigned int word, unsigned char *data);
extern int decode_frame(const unsigned char *buff, eph_t *eph, double *ion,
                        double *utc, int *leaps);

extern int init_raw   (raw_t *raw);
extern void free_raw  (raw_t *raw);
extern int input_raw  (raw_t *raw, int format, unsigned char data);
extern int input_rawf (raw_t *raw, int format, FILE *fp);

extern int input_oem4 (raw_t *raw, unsigned char data);
extern int input_oem3 (raw_t *raw, unsigned char data);
extern int input_ubx  (raw_t *raw, unsigned char data);
extern int input_ss2  (raw_t *raw, unsigned char data);
extern int input_cres (raw_t *raw, unsigned char data);
extern int input_oem4f(raw_t *raw, FILE *fp);
extern int input_oem3f(raw_t *raw, FILE *fp);
extern int input_ubxf (raw_t *raw, FILE *fp);
extern int input_ss2f (raw_t *raw, FILE *fp);
extern int input_cresf(raw_t *raw, FILE *fp);

/* rtcm functions ------------------------------------------------------------*/
extern int init_rtcm   (rtcm_t *rtcm);
extern void free_rtcm  (rtcm_t *rtcm);
extern int input_rtcm2 (rtcm_t *rtcm, unsigned char data);
extern int input_rtcm3 (rtcm_t *rtcm, unsigned char data);
extern int input_rtcm2f(rtcm_t *rtcm, FILE *fp);
extern int input_rtcm3f(rtcm_t *rtcm, FILE *fp);

/* solution functions --------------------------------------------------------*/
extern int readsol (char *files[], int nfile, solp_t *solp, double rpos[][3],
                    int *nrpos, char **inputs, int *ninp);
extern int readsolt(char *files[], int nfile, gtime_t ts, gtime_t te,
                    double tint, int qflag, solp_t *solp, double rpos[][3],
                    int *nrpos, char **inputs, int *ninp);
extern int outsolheads(unsigned char *buff);
extern int outsols(unsigned char *buff, gtime_t time, const double *rr,
                   const double *Qr, const double *rb, int stat, int ns,
                   double age, double ratio);
extern int outsolexs(unsigned char *buff, int stat, int ns, int *sat,
                     double *az, double *el, int *snr, int *vsat);
extern void outsolhead(FILE *fp);
extern void outsol(FILE *fp, gtime_t time, const double *rr, const double *Qr,
                   const double *rb, int stat, int ns, double age, double ratio);
extern void outsolex(FILE*fp, int stat, int ns, int *sat, double *az,
                     double *el, int *snr, int *vsat);
extern void setsolopt(const solopt_t *solopt);
extern void setsolformat(int format);

extern int outnmea_rmc(unsigned char *buff, gtime_t t, const double *rr,
                       int stat);
extern int outnmea_gga(unsigned char *buff, gtime_t t, const double *rr,
                       int stat, int ns, double age);
extern int outnmea_gsa(unsigned char *buff, int stat, int ns, int *sat,
                       double *az, double *el, int *snr, int *vsat);
extern int outnmea_gsv(unsigned char *buff, int stat, int ns, int *sat,
                       double *az, double *el, int *snr, int *vsat);

/* convert solutions to google earth kml file --------------------------------*/
extern int convkml(const char *infile, const char *outfile, gtime_t ts,
                   gtime_t te, double tint, int qflg, double *offset,
                   int tcolor, int pcolor, int outalt, int outtime);

/* SBAS functions ------------------------------------------------------------*/
extern int  sbsreadmsg (const char *file, int sel, sbsmsg_t **sbsmsg);
extern int  sbsreadmsgt(const char *file, int sel, gtime_t ts, gtime_t te,
                       sbsmsg_t **sbsmsg);
extern void sbsoutmsg(FILE *fp, sbsmsg_t *sbsmsg);
extern void sbsupdatestat(const sbsmsg_t *msg, sbsstat_t *stat);
extern int  sbsdecodemsg(gtime_t time, int prn, const unsigned int *words,
                         sbsmsg_t *sbsmsg);
extern int  sbssatpos(gtime_t time, const sbsnav_t *nav, double pr, double *rs,
                      double *dts);
extern int  sbspntpos(const obsd_t *obs, int n, const nav_t *nav,
                      const sbsstat_t *stat, double elmin, double snrmin,
                      int opt, double *rr, double *Qr, double *dtr,
                      double *azel, int *vsat, int *corr);

/* integer least-square estimation -------------------------------------------*/
extern int lambda(int n, int m, const double *a, const double *Q, double *F,
                  double *s);

/* realtime kinematic positioning --------------------------------------------*/
extern void rtkinit(rtk_t *rtk, const prcopt_t *opt);
extern void rtkfree(rtk_t *rtk);
extern int  rtkpos (rtk_t *rtk, const obsd_t *obs, int nobs, const nav_t *nav,
                    const sbsstat_t *sbsstat);

/* post-processing positioning -----------------------------------------------*/
extern int postpos(gtime_t ts, gtime_t te, double tint, int dirs,
                   char **infile, int nfile, char *outfile, int *informat);
extern void postposopt(const prcopt_t *prcopt_, const solopt_t *solopt_,
                       const char *ephfile_, const char *sapfile_);
extern int readopts (const char *file, prcopt_t *prcopt, solopt_t *solopt);
extern int writeopts(const char *file, const prcopt_t *prcopt,
                     const solopt_t *solopt);

/* stream data input/output --------------------------------------------------*/
extern void strinitcom(void);
extern void strinit  (stream_t *stream);
extern void strlock  (stream_t *stream);
extern void strunlock(stream_t *stream);
extern int  stropen  (stream_t *stream, int type, int mode, const char *path);
extern void strclose (stream_t *stream);
extern int  strread  (stream_t *stream, unsigned char *buff, int n);
extern int  strwrite (stream_t *stream, unsigned char *buff, int n);
extern void strsync  (stream_t *stream1, stream_t *stream2);
extern int  strstat  (stream_t *stream, char *msg);
extern void strsum   (stream_t *stream, int *inb, int *inr, int *outb, int *outr);
extern void strsetopt(const int *opt);
extern gtime_t strgettime(stream_t *stream);

/* stream server functions ---------------------------------------------------*/
extern void strsvrinit (strsvr_t *svr, int nout);
extern int  strsvrstart(strsvr_t *svr, int *opts, int *strs, char **paths,
                        const char *cmd, const double *nmeapos);
extern void strsvrstop (strsvr_t *svr, const char *cmd);
extern void strsvrstat (strsvr_t *svr, int *stat, int *byte, int *bps, char *msg);

/* rtk server functions ------------------------------------------------------*/
extern int  rtksvrinit  (rtksvr_t *svr);
extern int  rtksvrstart (rtksvr_t *svr, int cycle, int buffsize, int *strs,
                         char **paths, int *formats, int navsel, int sbssat,
                         char **cmds, int nmeacycle, int nmeareq,
                         const double *nmeapos, prcopt_t *prcopt,
                         solopt_t *solopt);
extern void rtksvrstop  (rtksvr_t *svr, char **cmds);
extern void rtksvrlock  (rtksvr_t *svr);
extern void rtksvrunlock(rtksvr_t *svr);
extern int  rtksvrostat (rtksvr_t *svr, int type, gtime_t *time, int *sat,
                         double *az, double *el, int *snr1, int *snr2, int *vsat);
extern void rtksvrsstat (rtksvr_t *svr, int *sstat, char *msg);

#ifdef __cplusplus
}
#endif
#endif /* RTKLIB_H */
